@@ -2,23 +2,26 @@
  * Copyright © 2008 Kristian Høgsberg
  * Copyright © 2013 Jason Ekstrand
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #define _GNU_SOURCE
@@ -30,7 +33,6 @@
 #include <stdio.h>
 #include <errno.h>
 #include <sys/uio.h>
-#include <assert.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -161,10 +163,10 @@ wl_connection_create(int fd)
 {
 	struct wl_connection *connection;
 
-	connection = malloc(sizeof *connection);
+	connection = zalloc(sizeof *connection);
 	if (connection == NULL)
 		return NULL;
-	memset(connection, 0, sizeof *connection);
+
 	connection->fd = fd;
 
 	return connection;
@@ -189,13 +191,16 @@ close_fds(struct wl_buffer *buffer, int max)
 	buffer->tail += size;
 }
 
-void
+int
 wl_connection_destroy(struct wl_connection *connection)
 {
+	int fd = connection->fd;
+
 	close_fds(&connection->fds_out, -1);
 	close_fds(&connection->fds_in, -1);
-	close(connection->fd);
 	free(connection);
+
+	return fd;
 }
 
 void
@@ -287,7 +292,7 @@ wl_connection_flush(struct wl_connection *connection)
 		msg.msg_namelen = 0;
 		msg.msg_iov = iov;
 		msg.msg_iovlen = count;
-		msg.msg_control = cmsg;
+		msg.msg_control = (clen > 0) ? cmsg : NULL;
 		msg.msg_controllen = clen;
 		msg.msg_flags = 0;
 
@@ -309,6 +314,12 @@ wl_connection_flush(struct wl_connection *connection)
 	return connection->out.head - tail;
 }
 
+uint32_t
+wl_connection_pending_input(struct wl_connection *connection)
+{
+	return wl_buffer_size(&connection->in);
+}
+
 int
 wl_connection_read(struct wl_connection *connection)
 {
@@ -345,7 +356,7 @@ wl_connection_read(struct wl_connection *connection)
 
 	connection->in.head += len;
 
-	return connection->in.head - connection->in.tail;
+	return wl_connection_pending_input(connection);
 }
 
 int
@@ -394,6 +405,12 @@ wl_message_count_arrays(const struct wl_message *message)
 	return arrays;
 }
 
+int
+wl_connection_get_fd(struct wl_connection *connection)
+{
+	return connection->fd;
+}
+
 static int
 wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 {
@@ -563,16 +580,12 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 		case 'h':
 			fd = args[i].h;
 			dup_fd = wl_os_dupfd_cloexec(fd, 0);
-			if (dup_fd < 0) {
-				wl_log("dup failed: %m");
-				abort();
-			}
+			if (dup_fd < 0)
+				wl_abort("dup failed: %s\n", strerror(errno));
 			closure->args[i].h = dup_fd;
 			break;
 		default:
-			wl_log("unhandled format code: '%c'\n",
-				arg.type);
-			assert(0);
+			wl_abort("unhandled format code: '%c'\n", arg.type);
 			break;
 		}
 	}
@@ -765,8 +778,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].h = fd;
 			break;
 		default:
-			wl_log("unknown type\n");
-			assert(0);
+			wl_abort("unknown type\n");
 			break;
 		}
 	}
@@ -785,19 +797,6 @@ wl_connection_demarshal(struct wl_connection *connection,
 	return NULL;
 }
 
-int
-wl_interface_equal(const struct wl_interface *a, const struct wl_interface *b)
-{
-	/* In most cases the pointer equality test is sufficient.
-	 * However, in some cases, depending on how things are split
-	 * across shared objects, we can end up with multiple
-	 * instances of the interface metadata constants.  So if the
-	 * pointers match, the interfaces are equal, if they don't
-	 * match we have to compare the interface names. */
-
-	return a == b || strcmp(a->name, b->name) == 0;
-}
-
 int
 wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 {
@@ -826,7 +825,6 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 			} else if (object == NULL && id != 0) {
 				wl_log("unknown object (%u), message %s(%s)\n",
 				       id, message->name, message->signature);
-				object = NULL;
 				errno = EINVAL;
 				return -1;
 			}
@@ -900,8 +898,7 @@ convert_arguments_to_ffi(const char *signature, uint32_t flags,
 			ffi_args[i] = &args[i].h;
 			break;
 		default:
-			wl_log("unknown type\n");
-			assert(0);
+			wl_abort("unknown type\n");
 			break;
 		}
 	}
@@ -931,6 +928,10 @@ wl_closure_invoke(struct wl_closure *closure, uint32_t flags,
 		     count + 2, &ffi_type_void, ffi_types);
 
 	implementation = target->implementation;
+	if (!implementation[opcode]) {
+		wl_abort("listener function for opcode %u of %s is NULL\n",
+			 opcode, target->interface->name);
+	}
 	ffi_call(&cif, implementation[opcode], NULL, ffi_args);
 }
 
@@ -1122,7 +1123,7 @@ wl_closure_send(struct wl_closure *closure, struct wl_connection *connection)
 		return -1;
 
 	buffer_size = buffer_size_for_closure(closure);
-	buffer = malloc(buffer_size * sizeof buffer[0]);
+	buffer = zalloc(buffer_size * sizeof buffer[0]);
 	if (buffer == NULL)
 		return -1;
 
