@@ -24,9 +24,11 @@
  * SOFTWARE.
  */
 
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <stdint.h>
 #include <string.h>
 #include <assert.h>
 #include <sys/socket.h>
@@ -35,6 +37,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <pthread.h>
+#include <poll.h>
 
 #include "wayland-private.h"
 #include "wayland-server.h"
@@ -95,21 +98,53 @@ empty_client(void)
 TEST(tc_leaks_tests)
 {
 	struct display *d = display_create();
-	client_create(d, empty_client);
+	client_create_noarg(d, empty_client);
 	display_run(d);
 	display_destroy(d);
 }
 
+/* This is how pre proxy-version registry binds worked,
+ * this should create a proxy that shares the display's
+ * version number: 0 */
+static void *
+old_registry_bind(struct wl_registry *wl_registry,
+		  uint32_t name,
+		  const struct wl_interface *interface,
+		  uint32_t version)
+{
+	struct wl_proxy *id;
+
+	id = wl_proxy_marshal_constructor(
+		(struct wl_proxy *) wl_registry, WL_REGISTRY_BIND,
+		interface, name, interface->name, version, NULL);
+
+	return (void *) id;
+}
+
+struct handler_info {
+	struct wl_seat *seat;
+	uint32_t bind_version;
+	bool use_unversioned;
+};
+
 static void
 registry_handle_globals(void *data, struct wl_registry *registry,
 			uint32_t id, const char *intf, uint32_t ver)
 {
-	struct wl_seat **seat = data;
+	struct handler_info *hi = data;
+
+	/* This is only for the proxy version test */
+	if (hi->bind_version)
+		ver = hi->bind_version;
 
 	if (strcmp(intf, "wl_seat") == 0) {
-		*seat = wl_registry_bind(registry, id,
-					 &wl_seat_interface, ver);
-		assert(*seat);
+		if (hi->use_unversioned)
+			hi->seat = old_registry_bind(registry, id,
+						     &wl_seat_interface, ver);
+		else
+			hi->seat = wl_registry_bind(registry, id,
+						    &wl_seat_interface, ver);
+		assert(hi->seat);
 	}
 }
 
@@ -119,31 +154,40 @@ static const struct wl_registry_listener registry_listener = {
 };
 
 static struct wl_seat *
-client_get_seat(struct client *c)
+client_get_seat_with_info(struct client *c, struct handler_info *hi)
 {
-	struct wl_seat *seat;
 	struct wl_registry *reg = wl_display_get_registry(c->wl_display);
 	assert(reg);
 
-	wl_registry_add_listener(reg, &registry_listener, &seat);
+	assert(hi);
+	hi->seat = NULL;
+	wl_registry_add_listener(reg, &registry_listener, hi);
 	wl_display_roundtrip(c->wl_display);
-	assert(seat);
+	assert(hi->seat);
 
 	wl_registry_destroy(reg);
 
-	return seat;
+	return hi->seat;
+}
+
+static struct wl_seat *
+client_get_seat(struct client *c)
+{
+	struct handler_info hi;
+
+	hi.use_unversioned = false;
+	hi.bind_version = 0;
+
+	return client_get_seat_with_info(c, &hi);
 }
 
 static void
-check_for_error(struct client *c, struct wl_proxy *proxy)
+check_pending_error(struct client *c, struct wl_proxy *proxy)
 {
 	uint32_t ec, id;
 	int err;
 	const struct wl_interface *intf;
 
-	/* client should be disconnected */
-	assert(wl_display_roundtrip(c->wl_display) == -1);
-
 	err = wl_display_get_error(c->wl_display);
 	assert(err == EPROTO);
 
@@ -153,6 +197,28 @@ check_for_error(struct client *c, struct wl_proxy *proxy)
 	assert(id == wl_proxy_get_id(proxy));
 }
 
+static void
+check_for_error(struct client *c, struct wl_proxy *proxy)
+{
+	/* client should be disconnected */
+	assert(wl_display_roundtrip(c->wl_display) == -1);
+
+	check_pending_error(c, proxy);
+}
+
+static struct client_info *
+find_client_info(struct display *d, struct wl_client *client)
+{
+	struct client_info *ci;
+
+	wl_list_for_each(ci, &d->clients, link) {
+		if (ci->wl_client == client)
+			return ci;
+	}
+
+	return NULL;
+}
+
 static void
 bind_seat(struct wl_client *client, void *data,
 	  uint32_t vers, uint32_t id)
@@ -161,16 +227,14 @@ bind_seat(struct wl_client *client, void *data,
 	struct client_info *ci;
 	struct wl_resource *res;
 
-	/* find the right client_info struct and save the
-	 * resource as its data, so that we can use it later */
-	wl_list_for_each(ci, &d->clients, link) {
-		if (ci->wl_client == client)
-			break;
-	}
+	ci = find_client_info(d, client);
+	assert(ci);
 
 	res = wl_resource_create(client, &wl_seat_interface, vers, id);
 	assert(res);
 
+	/* save the resource as client's info data,
+	 * so that we can use it later */
 	ci->data = res;
 }
 
@@ -209,7 +273,7 @@ TEST(post_error_to_one_client)
 	wl_global_create(d->wl_display, &wl_seat_interface,
 			 1, d, bind_seat);
 
-	cl = client_create(d, post_error_main);
+	cl = client_create_noarg(d, post_error_main);
 	display_run(d);
 
 	/* the display was stopped by client, so it can
@@ -264,8 +328,8 @@ TEST(post_error_to_one_from_two_clients)
 	wl_global_create(d->wl_display, &wl_seat_interface,
 			 1, d, bind_seat);
 
-	client_create(d, post_error_main2);
-	cl = client_create(d, post_error_main3);
+	client_create_noarg(d, post_error_main2);
+	cl = client_create_noarg(d, post_error_main3);
 	display_run(d);
 
 	/* post error only to the second client */
@@ -289,8 +353,8 @@ TEST(post_error_to_two_clients)
 	wl_global_create(d->wl_display, &wl_seat_interface,
 			 1, d, bind_seat);
 
-	cl = client_create(d, post_error_main3);
-	cl2 = client_create(d, post_error_main3);
+	cl = client_create_noarg(d, post_error_main3);
+	cl2 = client_create_noarg(d, post_error_main3);
 
 	display_run(d);
 
@@ -331,7 +395,7 @@ TEST(post_nomem_tst)
 	wl_global_create(d->wl_display, &wl_seat_interface,
 			 1, d, bind_seat);
 
-	cl = client_create(d, post_nomem_main);
+	cl = client_create_noarg(d, post_nomem_main);
 	display_run(d);
 
 	assert(cl->data);
@@ -340,7 +404,7 @@ TEST(post_nomem_tst)
 
 	/* first client terminated. Run it again,
 	 * but post no memory to client */
-	cl = client_create(d, post_nomem_main);
+	cl = client_create_noarg(d, post_nomem_main);
 	display_run(d);
 
 	assert(cl->data);
@@ -447,7 +511,7 @@ TEST(threading_errors_tst)
 {
 	struct display *d = display_create();
 
-	client_create(d, threading_post_err);
+	client_create_noarg(d, threading_post_err);
 	display_run(d);
 
 	display_destroy(d);
@@ -502,7 +566,7 @@ TEST(threading_cancel_read_tst)
 {
 	struct display *d = display_create();
 
-	client_create(d, threading_cancel_read);
+	client_create_noarg(d, threading_cancel_read);
 	display_run(d);
 
 	display_destroy(d);
@@ -542,7 +606,7 @@ threading_read_eagain(void)
 TEST(threading_read_eagain_tst)
 {
 	struct display *d = display_create();
-	client_create(d, threading_read_eagain);
+	client_create_noarg(d, threading_read_eagain);
 
 	display_run(d);
 
@@ -604,8 +668,261 @@ TEST(threading_read_after_error_tst)
 {
 	struct display *d = display_create();
 
-	client_create(d, threading_read_after_error);
+	client_create_noarg(d, threading_read_after_error);
 	display_run(d);
 
 	display_destroy(d);
 }
+
+static void
+wait_for_error_using_dispatch(struct client *c, struct wl_proxy *proxy)
+{
+	int ret;
+
+	while (true) {
+		/* Dispatching should eventually hit the protocol error before
+		 * any other error. */
+		ret = wl_display_dispatch(c->wl_display);
+		if (ret == 0) {
+			continue;
+		} else {
+			assert(errno == EPROTO);
+			break;
+		}
+	}
+
+	check_pending_error(c, proxy);
+}
+
+static void
+wait_for_error_using_prepare_read(struct client *c, struct wl_proxy *proxy)
+{
+	int ret = 0;
+	struct pollfd pfd[2];
+
+	while (true) {
+		while (wl_display_prepare_read(c->wl_display) != 0 &&
+		      errno == EAGAIN) {
+			assert(wl_display_dispatch_pending(c->wl_display) >= 0);
+		}
+
+		/* Flush may fail due to EPIPE if the connection is broken, but
+		 * this must not set a fatal display error because that would
+		 * result in it being impossible to read a potential protocol
+		 * error. */
+		do {
+			ret = wl_display_flush(c->wl_display);
+		} while (ret == -1 && (errno == EINTR || errno == EAGAIN));
+		assert(ret >= 0 || errno == EPIPE);
+		assert(wl_display_get_error(c->wl_display) == 0);
+
+		pfd[0].fd = wl_display_get_fd(c->wl_display);
+		pfd[0].events = POLLIN;
+		do {
+			ret = poll(pfd, 1, -1);
+		} while (ret == -1 && errno == EINTR);
+		assert(ret != -1);
+
+		/* We should always manage to read the error before the EPIPE
+		 * comes this way. */
+		assert(wl_display_read_events(c->wl_display) == 0);
+
+		/* Dispatching should eventually hit the protocol error before
+		 * any other error. */
+		ret = wl_display_dispatch_pending(c->wl_display);
+		if (ret == 0) {
+			continue;
+		} else {
+			assert(errno == EPROTO);
+			break;
+		}
+	}
+
+	check_pending_error(c, proxy);
+}
+
+static void
+check_error_after_epipe(void *data)
+{
+	bool use_dispatch_helpers = *(bool *) data;
+	struct client *client;
+	struct wl_seat *seat;
+	struct wl_callback *callback;
+
+	client = client_connect();
+
+	/* This will, according to the implementation below, cause the server
+	 * to post an error. */
+	seat = client_get_seat(client);
+	wl_display_flush(client->wl_display);
+
+	/* The server will not actually destroy the client until it receives
+	 * input, so send something to trigger the client destruction. */
+	callback = wl_display_sync(client->wl_display);
+	wl_callback_destroy(callback);
+
+	/* Sleep some to give the server a chance to react and destroy the
+	 * client. */
+	test_usleep(200000);
+
+	/* Wait for the protocol error and check that we reached it before
+	 * EPIPE. */
+	if (use_dispatch_helpers) {
+		wait_for_error_using_dispatch(client, (struct wl_proxy *) seat);
+	} else {
+		wait_for_error_using_prepare_read(client,
+						  (struct wl_proxy *) seat);
+	}
+
+	wl_seat_destroy(seat);
+	client_disconnect_nocheck(client);
+}
+
+static void
+bind_seat_and_post_error(struct wl_client *client, void *data,
+			 uint32_t version, uint32_t id)
+{
+	struct display *d = data;
+	struct client_info *ci;
+	struct wl_resource *resource;
+
+	ci = find_client_info(d, client);
+	assert(ci);
+
+	resource = wl_resource_create(client, &wl_seat_interface, version, id);
+	assert(resource);
+	ci->data = resource;
+
+	wl_resource_post_error(ci->data, 23, "Dummy error");
+}
+
+TEST(error_code_after_epipe)
+{
+	struct display *d = display_create();
+	bool use_dispatch_helpers;
+
+	wl_global_create(d->wl_display, &wl_seat_interface,
+			 1, d, bind_seat_and_post_error);
+
+	use_dispatch_helpers = true;
+	client_create(d, check_error_after_epipe, &use_dispatch_helpers);
+	display_run(d);
+
+	use_dispatch_helpers = false;
+	client_create(d, check_error_after_epipe, &use_dispatch_helpers);
+	display_run(d);
+
+	display_destroy(d);
+}
+
+static void
+check_seat_versions(struct wl_seat *seat, uint32_t ev)
+{
+	struct wl_pointer *pointer;
+
+	assert(wl_proxy_get_version((struct wl_proxy *) seat) == ev);
+	assert(wl_seat_get_version(seat) == ev);
+
+	pointer = wl_seat_get_pointer(seat);
+	assert(wl_pointer_get_version(pointer) == ev);
+	assert(wl_proxy_get_version((struct wl_proxy *) pointer) == ev);
+	wl_proxy_destroy((struct wl_proxy *) pointer);
+}
+
+/* Normal client with proxy versions available. */
+static void
+seat_version(void *data)
+{
+	struct handler_info *hi = data;
+	struct client *c = client_connect();
+	struct wl_seat *seat;
+
+	/* display proxy should always be version 0 */
+	assert(wl_proxy_get_version((struct wl_proxy *) c->wl_display) == 0);
+
+	seat = client_get_seat_with_info(c, hi);
+	if (hi->use_unversioned)
+		check_seat_versions(seat, 0);
+	else
+		check_seat_versions(seat, hi->bind_version);
+
+	wl_proxy_destroy((struct wl_proxy *) seat);
+
+	client_disconnect_nocheck(c);
+}
+
+TEST(versions)
+{
+	struct display *d = display_create();
+	struct wl_global *global;
+	int i;
+
+	global = wl_global_create(d->wl_display, &wl_seat_interface,
+				  5, d, bind_seat);
+
+	for (i = 1; i <= 5; i++) {
+		struct handler_info hi;
+
+		hi.bind_version = i;
+		hi.use_unversioned = false;
+		client_create(d, seat_version, &hi);
+		hi.use_unversioned = true;
+		client_create(d, seat_version, &hi);
+	}
+
+	display_run(d);
+
+	wl_global_destroy(global);
+
+	display_destroy(d);
+}
+
+static void
+check_error_on_destroyed_object(void *data)
+{
+	struct client *c;
+	struct wl_seat *seat;
+	uint32_t id;
+	const struct wl_interface *intf;
+
+	c = client_connect();
+	seat = client_get_seat(c);
+
+	/* destroy the seat proxy. The display won't know
+	 * about it yet, so it will post the error as usual */
+	wl_proxy_destroy((struct wl_proxy *) seat);
+
+	/* let display post the error. The error will
+	 * be caught in stop_display while dispatching */
+	assert(stop_display(c, 1) == -1);
+
+	/* check the returned error. Since the object was destroyed,
+	 * we don't know the interface and id */
+	assert(wl_display_get_error(c->wl_display) == EPROTO);
+	assert(wl_display_get_protocol_error(c->wl_display, &intf, &id) == 23);
+	assert(intf == NULL);
+	assert(id == 0);
+
+	client_disconnect_nocheck(c);
+}
+
+TEST(error_on_destroyed_object)
+{
+	struct client_info *cl;
+	struct display *d = display_create();
+
+	wl_global_create(d->wl_display, &wl_seat_interface,
+			 1, d, bind_seat);
+
+	cl = client_create_noarg(d, check_error_on_destroyed_object);
+	display_run(d);
+
+	/* did client bind to the seat? */
+	assert(cl->data);
+
+	/* post error on the destroyed object */
+	wl_resource_post_error((struct wl_resource *) cl->data,
+			       23, "Dummy error");
+	display_resume(d);
+	display_destroy(d);
+}
